""".. Ignore pydocstyle D400.

=========================================================================
Collect Processes' tools and store them in configmap object in Kubernetes
=========================================================================

"""
import hashlib
import json
import logging
from pathlib import Path
from typing import Any, Dict

import kubernetes

from django.conf import settings
from django.core.management.base import BaseCommand, CommandError

from resolwe.flow.managers.workload_connectors.kubernetes import (
    sanitize_kubernetes_label,
)
from resolwe.flow.utils import get_apps_tools
from resolwe.utils import BraceMessage as __

# Timeout (in seconds) to wait for response from kubernetes API.
KUBERNETES_TIMEOUT = 30
KUBERNETES_NAMESPACE = getattr(settings, "KUBERNETES_SETTINGS", {}).get(
    "namespace", "default"
)

logger = logging.getLogger(__name__)


class Command(BaseCommand):
    """Copies tools' files from different locations into Configmap object.

    Multiple configmaps are created and
    """

    help = "Collect tools' files and store them as Configmap in Kubernetes."

    def add_arguments(self, parser):
        """Command arguments."""
        parser.add_argument(
            "--noinput",
            "--no-input",
            action="store_false",
            dest="interactive",
            help="Do NOT prompt the user for input of any kind.",
        )

    def get_confirmation(self):
        """Get user confirmation to proceed."""
        message = (
            "\n"
            "This will RECREATE all configmaps describing tools.\n"
            "Are you sure you want to do this?\n"
            "\n"
            "Type 'yes' to continue, or 'no' to cancel."
        )

        if input("".join(message)) != "yes":
            raise CommandError("Collecting tools cancelled.")

    def update_configmap(self, name: str, content: Dict, core_api: Any):
        """Updata/create configmap."""
        meta = kubernetes.client.V1ObjectMeta(name=name, namespace=KUBERNETES_NAMESPACE)
        configmap = kubernetes.client.V1ConfigMap(
            data=content, metadata=meta, kind="ConfigMap"
        )
        try:
            core_api.replace_namespaced_config_map(
                name=name, namespace=KUBERNETES_NAMESPACE, body=configmap
            )
        except kubernetes.client.rest.ApiException:
            # The configmap was not found, create it.
            core_api.create_namespaced_config_map(
                namespace=KUBERNETES_NAMESPACE, body=configmap
            )

    def update_tools_configmaps(self, core_api):
        """Create or update configmaps for tools."""

        def dict_from_directory(directory: Path) -> Dict[str, str]:
            """Get dictionary from given directory.

            File names are keys and corresponding file contents are values.
            """
            return {
                entry.name: entry.read_text()
                for entry in directory.glob("*")
                if entry.is_file()
            }

        configmaps = dict()
        for app_name, tool_path in get_apps_tools().items():
            logger.info(__("Processing '{}' from '{}'.", app_name, tool_path))
            data = dict_from_directory(tool_path)
            data_md5 = hashlib.md5(
                json.dumps(data, sort_keys=True).encode()
            ).hexdigest()
            configmap_name = sanitize_kubernetes_label(f"tools-{app_name}-{data_md5}")
            logger.info(__("Assigned configmap name '{}'.", configmap_name))
            self.update_configmap(configmap_name, data, core_api)
            configmaps[sanitize_kubernetes_label(app_name)] = configmap_name

        description_configmap_name = getattr(
            settings, "KUBERNETES_TOOLS_CONFIGMAPS", "tools-configmaps"
        )
        logger.info(__("Updating main configmap '{}'", description_configmap_name))
        self.update_configmap(description_configmap_name, configmaps, core_api)

    def handle(self, **options):
        """Collect tools."""
        try:
            kubernetes.config.load_kube_config()
        except kubernetes.config.config_exception.ConfigException:
            kubernetes.config.load_incluster_config()

        core_api = kubernetes.client.CoreV1Api()

        if options["interactive"]:
            self.get_confirmation()

        self.update_tools_configmaps(core_api=core_api)



# Generated by Django 3.2.19 on 2023-06-06 09:22

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("flow", "0009_preset_base_model"),
    ]

    operations = [
        migrations.AddField(
            model_name="annotationfield",
            name="required",
            field=models.BooleanField(default=False),
        ),
    ]



"""Resolwe storage serializer."""
from resolwe.flow.models import Storage
from resolwe.rest.fields import ProjectableJSONField

from .base import ResolweBaseSerializer


class StorageSerializer(ResolweBaseSerializer):
    """Serializer for Storage objects."""

    json = ProjectableJSONField()

    class Meta:
        """StorageSerializer Meta options."""

        model = Storage
        read_only_fields = (
            "created",
            "id",
            "modified",
        )
        update_protected_fields = ("contributor",)
        fields = (
            read_only_fields
            + update_protected_fields
            + (
                "data",
                "json",
                "name",
                "slug",
            )
        )



"""Basic listener commands."""
import logging
import re
from collections import defaultdict
from contextlib import suppress
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Union

from django.conf import settings
from django.db import transaction
from django.db.models import Sum
from django.db.models.functions import Coalesce
from django.urls import reverse
from django.utils.timezone import now

from resolwe.flow.executors.socket_utils import Message, Response
from resolwe.flow.managers.protocol import ExecutorProtocol
from resolwe.flow.models import Data, DataDependency, Entity, Process, Worker
from resolwe.flow.models.utils import validate_data_object
from resolwe.flow.utils import dict_dot, iterate_fields, iterate_schema
from resolwe.storage.connectors import connectors
from resolwe.storage.connectors.hasher import StreamHasher
from resolwe.storage.models import ReferencedPath, StorageLocation
from resolwe.utils import BraceMessage as __

from .plugin import ListenerPlugin, listener_plugin_manager

if TYPE_CHECKING:
    from resolwe.flow.managers.listener.listener import Processor

logger = logging.getLogger(__name__)


class BasicCommands(ListenerPlugin):
    """Basic listener handlers."""

    plugin_manager = listener_plugin_manager

    def handle_run(
        self, data_id: int, message: Message[dict], manager: "Processor"
    ) -> Response[int]:
        """Handle spawning new data object.

        The response is the id of the created object or -1 when object can not
        be created.
        """
        export_files_mapper = message.message_data["export_files_mapper"]
        processing_data: Data = Data.objects.get(pk=data_id)

        try:
            data = message.message_data["data"]
            logger.debug(__("Spawning new data object from dict: {}", data))

            data["contributor"] = processing_data.contributor
            data["process"] = (
                Process.objects.filter(slug=data["process"])
                .filter_for_user(processing_data.contributor)
                .latest()
            )
            data["tags"] = processing_data.tags
            data["collection"] = processing_data.collection
            data["subprocess_parent"] = processing_data

            with transaction.atomic():
                for field_schema, fields in iterate_fields(
                    data.get("input", {}), data["process"].input_schema
                ):
                    type_ = field_schema["type"]
                    name = field_schema["name"]
                    value = fields[name]

                    if type_ == "basic:file:":
                        fields[name] = self.hydrate_spawned_files(
                            export_files_mapper, value
                        )
                    elif type_ == "list:basic:file:":
                        fields[name] = [
                            self.hydrate_spawned_files(export_files_mapper, fn)
                            for fn in value
                        ]
                created_object = Data.objects.create(**data)
        except Exception:
            manager._log_exception(
                processing_data,
                f"Error while preparing spawned Data objects for process '{processing_data.process.slug}'",
            )
            return message.respond_error(-1)
        else:
            return message.respond_ok(created_object.id)

    def handle_update_rc(
        self, data_id: int, message: Message[dict], manager: "Processor"
    ) -> Response[str]:
        """Update return code.

        When return code is chaned from zero to non-zero value the Data object
        status is set to ERROR.

        :args: data contains at least key rc and possibly key error.
        """
        data = manager.data(data_id)
        new_return_code = message.message_data["rc"]
        if new_return_code not in (0, data.process_rc):
            changes = ["process_rc"]
            if data.process_rc is None or data.process_rc == 0:
                max_length = Data._meta.get_field("process_error").base_field.max_length
                error_details = message.message_data.get("error", "Unspecified error")
                if len(error_details) > max_length:
                    error_details = error_details[: max_length - 3] + "..."
                # Avoid duplicate entries in process_error.
                if error_details not in data.process_error:
                    data.process_error.append(error_details)
                data.status = Data.STATUS_ERROR
                changes += ["process_error", "status"]
            data.process_rc = new_return_code
            manager._save_data(data, changes)
        return message.respond_ok("OK")

    def handle_finish(
        self, data_id: int, message: Message[Dict], manager: "Processor"
    ) -> Response[str]:
        """Handle an incoming ``Data`` finished processing request."""
        data = manager.data(data_id)
        process_rc = int(message.message_data.get("rc", 1))
        data.process_progress = 100
        data.finished = now()
        changes = ["process_progress", "finished", "size"]
        if process_rc != 0:
            data.process_rc = process_rc
            changes.append("process_rc")
            if data.status != Data.STATUS_ERROR:
                data.status = Data.STATUS_ERROR
                changes.append("status")
                error_message = message.message_data.get(
                    "error", "Process return code is not 0."
                )
                if error_message not in data.process_error:
                    data.process_error.append(error_message)
                    changes.append("process_error")
        elif data.status != Data.STATUS_ERROR:
            data.status = Data.STATUS_DONE
            changes.append("status")

        data.size = data.location.files.aggregate(size=Coalesce(Sum("size"), 0)).get(
            "size"
        )
        with transaction.atomic():
            manager._save_data(data, changes)
            data.worker.status = Worker.STATUS_COMPLETED
            data.worker.save()
            default_location = data.location.default_storage_location
            default_location.status = StorageLocation.STATUS_DONE
            default_location.save(update_fields=["status"])

        # Only validate objects with DONE status. Validating objects in ERROR
        # status will only cause unnecessary errors to be displayed.
        if data.status == Data.STATUS_DONE:
            validate_data_object(data)
        return message.respond_ok("OK")

    def handle_missing_data_locations(
        self, data_id: int, message: Message, manager: "Processor"
    ) -> Response[Union[str, Dict[str, Dict[str, Any]]]]:
        """Handle an incoming request to get missing data locations."""
        storage_name = "data"
        filesystem_connector = None
        filesystem_connectors = [
            connector
            for connector in connectors.for_storage(storage_name)
            if connector.mountable
        ]

        if filesystem_connectors:
            filesystem_connector = filesystem_connectors[0]

        missing_data = dict()
        dependencies = (
            Data.objects.filter(
                children_dependency__child_id=data_id,
                children_dependency__kind=DataDependency.KIND_IO,
            )
            .exclude(location__isnull=True)
            .exclude(pk=data_id)
            .distinct()
        )

        for parent in dependencies:
            file_storage = parent.location
            # Is location available on some local connector?
            if any(
                file_storage.has_storage_location(filesystem_connector.name)
                for filesystem_connector in filesystem_connectors
            ):
                continue

            from_location = file_storage.default_storage_location
            if from_location is None:
                data = manager.data(data_id)
                manager._log_exception(
                    data,
                    "No storage location exists (handle_get_missing_data_locations).",
                    extra={"file_storage_id": file_storage.id},
                )
                return message.respond_error("No storage location exists")

            # When there exists at least one filesystem connector for the data
            # storage download inputs to the shared storage.
            missing_data_item = {
                "data_id": parent.pk,
                "from_connector": from_location.connector_name,
            }
            if filesystem_connector:
                to_location = StorageLocation.all_objects.get_or_create(
                    file_storage=file_storage,
                    url=from_location.url,
                    connector_name=filesystem_connector.name,
                )[0]
                missing_data_item["from_storage_location_id"] = from_location.id
                missing_data_item["to_storage_location_id"] = to_location.id
                missing_data_item["to_connector"] = filesystem_connector.name
            else:
                missing_data_item["files"] = list(
                    ReferencedPath.objects.filter(
                        storage_locations=from_location
                    ).values()
                )

            missing_data[from_location.url] = missing_data_item
            # Set last modified time so it does not get deleted.
            from_location.last_update = now()
            from_location.save()
        return message.respond_ok(missing_data)

    def handle_referenced_files(
        self, data_id: int, message: Message[List[dict]], manager: "Processor"
    ) -> Response[str]:
        """Store  a list of files and directories produced by the worker.

        The files are a list of dictionaries with keys 'path', 'size' and
        hashes.
        """
        data = Data.objects.get(pk=data_id)
        update_fields = ["size"] + StreamHasher.KNOWN_HASH_TYPES
        storage_location = data.location.default_storage_location
        referenced_files = message.message_data
        paths = {
            referenced_file["path"]: referenced_file
            for referenced_file in referenced_files
        }
        updated_paths = set()
        with transaction.atomic():
            # Bulk update existing paths.
            database_paths = ReferencedPath.objects.filter(
                storage_locations=storage_location, path__in=paths.keys()
            )
            for database_path in database_paths:
                updated_paths.add(database_path.path)
                for key, value in paths[database_path.path].items():
                    setattr(database_path, key, value)
            ReferencedPath.objects.bulk_update(database_paths, update_fields)
            logger.debug("Updated %s.", database_paths.values())

            # Bulk insert new paths.
            created_paths = [
                ReferencedPath(**referenced_file)
                for referenced_file in referenced_files
                if referenced_file["path"] not in updated_paths
            ]
            ReferencedPath.objects.bulk_create(created_paths)
            storage_location.files.add(*created_paths)
            logger.debug("Created %s.", [e.path for e in created_paths])

        return message.respond_ok("OK")

    def handle_resolve_url(
        self, data_id: int, message: Message[str], manager: "Processor"
    ) -> Response[str]:
        """Resolve the download link.

        The download link is of the format protocol://link.
        Currently we support the protocol "s3".
        """

        def handle_s3(url: str) -> str:
            """Handle the S3 url.

            The URL is of the form s3://bucket-name/key_name.
            """
            # Avoid optional boto3 requirement on startup.
            from resolwe.storage.connectors.s3connector import AwsS3Connector

            allowed_characters = r"[1-9a-z\-\.]{3,63}"
            regexp = rf"^s3://(?P<bucket>{allowed_characters})/(?P<key>(.*){{1,1024}})$"
            with suppress(Exception):
                match = re.match(regexp, url)
                bucket, key = match["bucket"], match["key"]
                for connector in connectors.values():
                    if isinstance(connector, AwsS3Connector):
                        if connector.bucket_name == bucket:
                            # The generated URL can expire quite fast, since the
                            # download is triggered immediatelly.
                            presigned = connector.presigned_url(key, expiration=300)
                            if presigned is not None:
                                return presigned
            return url

        handle_default = lambda url: url
        url = message.message_data
        handlers: Dict[str, Callable[[str], str]] = defaultdict(lambda: handle_default)
        handlers["s3"] = handle_s3
        protocol_regex = r"^(?P<protocol>[a-zA-Z1-9]+)://"
        with suppress(Exception):
            protocol = re.match(protocol_regex, url)["protocol"].lower()
            return message.respond_ok(handlers[protocol](url))

        return message.respond_ok(url)

    def hydrate_spawned_files(
        self, exported_files_mapper: Dict[str, str], filename: str
    ):
        """Pop the given file's map from the exported files mapping.

        :param exported_files_mapper: The dict of file mappings this
            process produced.
        :param filename: The filename to format and remove from the
            mapping.
        :return: The formatted mapping between the filename and
            temporary file path.
        :rtype: dict
        """
        if filename not in exported_files_mapper:
            raise KeyError(
                "Use 're-export' to prepare the file for spawned process: {}".format(
                    filename
                )
            )
        return {"file_temp": exported_files_mapper.pop(filename), "file": filename}

    def handle_update_status(
        self, data_id: int, message: Message[str], manager: "Processor"
    ) -> Response[str]:
        """Update data status."""

        data_status = manager.get_data_fields(data_id, "status")
        new_status = manager._choose_worst_status(message.message_data, data_status)
        if new_status != data_status:
            data = manager.data(data_id)
            data.status = new_status

            with transaction.atomic():
                manager._save_data(data, ["status"])
                if new_status == Data.STATUS_PREPARING:
                    data.worker.status = Worker.STATUS_PREPARING
                    data.worker.save()
                elif new_status == Data.STATUS_PROCESSING:
                    data.worker.status = Worker.STATUS_PROCESSING
                    data.worker.save()

            if new_status == Data.STATUS_ERROR:
                logger.error(
                    __(
                        "Error occured while processing data object '{}' (handle_update).",
                        data_id,
                    ),
                    extra={
                        "data_id": data_id,
                        "api_url": "{}{}".format(
                            getattr(settings, "RESOLWE_HOST_URL", ""),
                            reverse(
                                "resolwe-api:data-detail",
                                kwargs={"pk": data_id},
                            ),
                        ),
                    },
                )
        return message.respond_ok(new_status)

    def handle_get_data_by_slug(
        self, data_id: int, message: Message[str], manager: "Processor"
    ) -> Response[int]:
        """Get data id by slug."""
        return message.respond_ok(
            Data.objects.filter(slug=message.message_data)
            .values_list("id", flat=True)
            .get()
        )

    def handle_set_data_size(
        self, data_id: int, message: Message[int], manager: "Processor"
    ) -> Response[str]:
        """Set size of data object.

        Note: the signals on data object are not triggered on size change.
        """
        assert isinstance(message.message_data, int)
        manager._update_data(data_id, {"size": message.message_data})
        return message.respond_ok("OK")

    def handle_update_output(
        self, data_id: int, message: Message[Dict[str, Any]], manager: "Processor"
    ) -> Response[str]:
        """Update data output."""
        data = manager.data(data_id)
        output_schema = manager.get_data_fields(data_id, ("process__output_schema"))
        storage_fields = {
            field_name
            for schema, _, field_name in iterate_schema(data.output, output_schema)
            if schema["type"].startswith("basic:json:")
        }

        with transaction.atomic():
            for key, val in message.message_data.items():
                if key in storage_fields:
                    val = manager.save_storage(key, val, data).pk
                dict_dot(data.output, key, val)
            manager._save_data(data, ["output"])
        return message.respond_ok("OK")

    def handle_get_files_to_download(
        self, data_id: int, message: Message[int], manager: "Processor"
    ) -> Response[Union[str, List[dict]]]:
        """Get a list of files belonging to a given storage location object."""
        return message.respond_ok(
            list(
                ReferencedPath.objects.filter(
                    storage_locations=message.message_data
                ).values()
            )
        )

    def handle_download_started(
        self, data_id: int, message: Message[dict], manager: "Processor"
    ) -> Response[str]:
        """Handle an incoming request to start downloading data.

        We have to check if the download for given StorageLocation object has
        already started.
        """
        storage_location_id = message.message_data[ExecutorProtocol.STORAGE_LOCATION_ID]
        lock = message.message_data.get(ExecutorProtocol.DOWNLOAD_STARTED_LOCK, False)
        with transaction.atomic():
            query = StorageLocation.all_objects.select_for_update().filter(
                pk=storage_location_id
            )
            location_status = query.values_list("status", flat=True).get()
            return_status = {
                StorageLocation.STATUS_PREPARING: ExecutorProtocol.DOWNLOAD_STARTED,
                StorageLocation.STATUS_UPLOADING: ExecutorProtocol.DOWNLOAD_IN_PROGRESS,
                StorageLocation.STATUS_DONE: ExecutorProtocol.DOWNLOAD_FINISHED,
            }[location_status]

            if location_status == StorageLocation.STATUS_PREPARING and lock:
                query.update(status=StorageLocation.STATUS_UPLOADING)

        return message.respond_ok(return_status)

    def handle_download_aborted(
        self, data_id: int, message: Message[int], manager: "Processor"
    ) -> Response[str]:
        """Handle an incoming download aborted request.

        Since download is aborted it does not matter if the location does
        exist or not so we do an update with only one SQL query.
        """
        StorageLocation.all_objects.filter(pk=message.message_data).update(
            status=StorageLocation.STATUS_PREPARING
        )
        return message.respond_ok("OK")

    def handle_download_finished(
        self, data_id: int, message: Message[int], manager: "Processor"
    ) -> Response[str]:
        """Handle an incoming download finished request."""
        storage_location = StorageLocation.all_objects.get(pk=message.message_data)
        # Add files to the downloaded storage location.
        default_storage_location = (
            storage_location.file_storage.default_storage_location
        )
        with transaction.atomic():
            storage_location.files.add(*default_storage_location.files.all())
            storage_location.status = StorageLocation.STATUS_DONE
            storage_location.save()
        return message.respond_ok("OK")

    def handle_annotate(
        self, data_id: int, message: Message[dict], manager: "Processor"
    ) -> Response[str]:
        """Handle an incoming ``Data`` object annotate request."""
        (entity_id,) = manager.get_data_fields(data_id, ["entity_id"])

        if entity_id is None:
            raise RuntimeError(f"No entity to annotate for object '{data_id}'")

        # The entity must be retrieved and saved or schema validation on the
        # entity must be done manually.
        entity = Entity.objects.get(pk=entity_id)
        for key, val in message.message_data.items():
            dict_dot(entity.descriptor, key, val)
        entity.save()

        return message.respond_ok("OK")

    def handle_process_log(
        self, data_id, message: Message[dict], manager: "Processor"
    ) -> Response[str]:
        """Handle an process log request."""
        data = manager.data(data_id)
        changes = []
        for key, values in message.message_data.items():
            data_key = f"process_{key}"
            changes.append(data_key)
            max_length = Data._meta.get_field(data_key).base_field.max_length
            if isinstance(values, str):
                values = [values]
            for i, entry in enumerate(values):
                if len(entry) > max_length:
                    values[i] = entry[: max_length - 3] + "..."
                # Avoid duplicates.
                log_list = getattr(data, data_key)
                if values[i] not in log_list:
                    log_list.append(values[i])
        if "process_error" in changes:
            data.status = Data.STATUS_ERROR
            changes.append("status")
        manager._save_data(data, changes)
        return message.respond_ok("OK")

    def handle_progress(
        self, data_id: int, message: Message[float], manager: "Processor"
    ) -> Response[str]:
        """Handle a progress change request.

        Do not trigger the signals when only progress changes.
        """
        manager._update_data(data_id, {"process_progress": message.message_data})
        return message.respond_ok("OK")



"""Admin pages."""
from django import forms
from django.contrib import admin
from django.contrib.admin import widgets
from django.contrib.auth.models import Group

from resolwe.flow.models import AnnotationField, AnnotationGroup, AnnotationPreset
from resolwe.permissions.models import Permission

admin.site.site_header = "Genialis Administration"
admin.site.site_title = "Genialis"


class AnnotationPresetForm(forms.ModelForm):
    """Form for AnnotationPreset admin page."""

    # Use FilteredSelectMultiple widget for filtering the annotation fields.
    fields = forms.ModelMultipleChoiceField(
        widget=widgets.FilteredSelectMultiple(
            verbose_name="Annotation fields", is_stacked=False
        ),
        queryset=AnnotationField.objects.all(),
    )

    # Groups with permission to view the preset.
    groups = forms.MultipleChoiceField(
        widget=widgets.FilteredSelectMultiple(verbose_name="Groups", is_stacked=False),
        # Required must be set to false to allow unselecting all entries.
        required=False,
    )

    class Meta:
        """Define model and fields for the form."""

        model = AnnotationPreset
        fields = ["name", "fields", "groups"]

    def __init__(self, *args, **kwargs):
        """Set the groups choices and initial value."""
        super().__init__(*args, **kwargs)
        group_choices = Group.objects.all().values_list("pk", "name").order_by("name")
        selected_group_ids = []
        if "instance" in kwargs:
            preset = kwargs["instance"]
            selected_group_ids = [
                group.pk for group in preset.groups_with_permission(Permission.VIEW)
            ]
        self.fields["groups"].choices = group_choices
        self.fields["groups"].initial = selected_group_ids

    def save(self, commit):
        """Save changes made to the preset."""
        preset = super().save(commit)
        # Remove all permissions on the preset.
        preset.permission_group.permissions.all().delete()
        # Add permission to selected groups.
        for group in Group.objects.filter(pk__in=self.cleaned_data["groups"]):
            preset.set_permission(Permission.VIEW, group)
        return preset


class AnnotationAdmin(admin.ModelAdmin):
    """Admin page for AnnotationPreset."""

    form = AnnotationPresetForm


admin.site.register(AnnotationField)
admin.site.register(AnnotationGroup)
admin.site.register(AnnotationPreset, AnnotationAdmin)



"""The redis cache for Django ORM."""

import logging
import pickle
import time
from contextlib import suppress
from datetime import datetime
from enum import Enum
from functools import partial
from itertools import islice
from os import getpid
from typing import Any, Iterable, Optional, Sequence, Union

import redis

from django.conf import settings
from django.db import models

from resolwe.flow.managers.listener.plugin_interface import Plugin, PluginManager
from resolwe.flow.models import Data
from resolwe.utils import BraceMessage as __

redis_server = redis.from_url(
    getattr(settings, "REDIS_CONNECTION_STRING", "redis://localhost")
)
logger = logging.getLogger(__name__)


class RedisLockStatus(Enum):
    """Redis lock status."""

    PROCESSING = "PR"
    OK = "OK"
    ERROR = "ER"


def chunked(iterable: Iterable, size: int):
    """Iterate the given iterable in chunks of the given size."""
    iterator = iter(iterable)
    while True:
        chunk = tuple(islice(iterator, size))
        if not chunk:
            break
        yield chunk


FieldNames = Sequence[str]
FieldValues = dict[str, Any]
Identifier = Sequence
Cache = dict[Identifier, FieldValues]


class RedisCache:
    """Stores a dictionaries with striLahkongs for keys.

    They are stored as hashes in Redis with pickle dumps as values.
    """

    # Max number of keys to retrieve from Redis in a single batch.
    KEY_BATCH_SIZE = 10000

    def __init__(self, *args, **kwargs):
        """Set the cached field set."""
        # The set of fields on the data object that can be safely cached inside
        # Redis to avoid hitting the database (we are assuming Redis is faster).
        # The list contains a set of commonly used fields that do not change
        # from the outside (or do no harm if they do).

        self._redis = redis_server
        super().__init__(*args, **kwargs)

    def _model_str(self, model: Optional[models.Model]) -> str:
        """Get string representation for the given content type."""
        return model._meta.label_lower if model else ""

    def _identifiers_str(self, identifiers: Optional[Sequence]) -> str:
        """Get string representation for the given identifier."""
        if not identifiers:
            return ""
        else:
            return "-".join(map(str, identifiers))

    def _cached_field_name(self, model: models.Model, field_name: str) -> str:
        """Get the cached field name."""
        return f"{self._model_str(model)}:{field_name}"

    def get_redis_key(self, model: models.Model, identifiers: Identifier) -> str:
        """Get the redis key from the field name.

        When testing every listener process must have its own redis shared
        storage since data ids will repeat when tests are run in parallel.
        """
        return self._get_redis_key_prefix(model, identifiers)

    def _get_redis_key_prefix(
        self,
        model: Optional[models.Model] = None,
        identifiers: Optional[Identifier] = None,
        field_name: Optional[str] = None,
    ) -> str:
        """Get redis key prefix based on input parameters.

        :raises RuntimeError: when input parameters are not valid.
        """
        if not model and (identifiers or field_name):
            raise RuntimeError(
                "When model is None then identifiers and field name must be provided."
            )
        if field_name and not identifiers:
            raise RuntimeError(
                "When field name is not None then identifiers must be provided."
            )

        from resolwe.test import is_testing  # Remove circular import.

        parts = [
            "listener" if not is_testing() else f"listener-{getpid()}",
            self._model_str(model),
            self._identifiers_str(identifiers),
            field_name,
        ]
        return ":".join([part for part in parts if part])

    def _get_redis_data(self, redis_keys: Sequence[str]) -> list[Optional[Any]]:
        """Retrieve the data from Redis for the given keys.

        The data is also unpickled, None is returned when data is not cached.
        """

        def unpickle(data: Optional[bytes]) -> Optional[FieldValues]:
            """Unpickle the data.

            When data is None return None.
            """
            return pickle.loads(data) if data is not None else None

        cached_data: list[Optional[FieldValues]] = list()
        for batch_keys in chunked(redis_keys, self.KEY_BATCH_SIZE):
            try:
                batch_data = self._redis.mget(*batch_keys)
            except redis.exceptions.RedisError:
                logger.exception(
                    __(
                        "Could not retrieve data from Redis for keys: '{}'.",
                        ", ".join(batch_keys),
                    )
                )
                raise

            try:
                cached_data.extend(map(unpickle, batch_data))
            except pickle.PickleError:
                logger.exception(
                    __(
                        "Could not deserialize data from Redis for keys: '{}'.",
                        batch_keys,
                    )
                )
                raise
        return cached_data  # type: ignore

    def mget(
        self, model: models.Model, identifiers_list: Sequence[Identifier]
    ) -> list[Optional[FieldValues]]:
        """Obtain the set of fields from the redis cache.

        All the cached field names are always returned.

        The query does not run in a transaction.

        :raises pickle.PickleError: when data cannot be unpickled.
        :raises redis.exceptions.RedisError: when data cannot be retrieved from Redis.
        """
        get_redis_key = partial(self.get_redis_key, model)
        redis_keys = list(map(get_redis_key, identifiers_list))
        # Wait for all keys to be unlocked or timeout occurs.
        # The refresh interval is set to 5 seconds since list can be long.
        self.wait(model, identifiers_list, timeout=60, refresh_interval=5)
        return self._get_redis_data(redis_keys)

    def _lock_key(self, model: models.Model, identifiers: Sequence) -> str:
        """Get the key for the lock for the given entry."""
        return self._get_redis_key_prefix(model, identifiers, "__lock__")

    def _modify_redis_locks(
        self,
        model: models.Model,
        identifiers_list: Sequence[Identifier],
        status: RedisLockStatus = RedisLockStatus.PROCESSING,
        valid_for: int = 60,
    ):
        """Modify the lock for the given entries.

        Create the lock indicating the given entry is processing. The lock will
        auto-expire after timeout seconds.
        """
        status_pickle = pickle.dumps(status)
        redis_map = {
            self._lock_key(model, identifier): status_pickle
            for identifier in identifiers_list
        }
        self._redis.mset(redis_map)
        for redis_key in redis_map:
            self._redis.expire(redis_key, valid_for)

    def lock(
        self,
        model: models.Model,
        identifiers_list: Sequence[Identifier],
        valid_for: int = 300,
    ):
        """Set the lock for the given entry.

        Create the lock indicating the given entry is processing. The lock will
        auto-expire after timeout seconds.
        """
        self._modify_redis_locks(
            model, identifiers_list, RedisLockStatus.PROCESSING, valid_for
        )

    def unlock(
        self,
        model: models.Model,
        identifiers_list: Sequence[Identifier],
        status: RedisLockStatus = RedisLockStatus.OK,
    ):
        """Release the lock for the given entry with status.

        :raise AssertionError: when status in not OK or ERROR.
        """
        assert status in (RedisLockStatus.OK, RedisLockStatus.ERROR)
        self._modify_redis_locks(model, identifiers_list, status)

    def _get_redis_locks(
        self, model: models.Model, identifiers_list: Sequence[Identifier]
    ) -> list[Optional[RedisLockStatus]]:
        """Get the locks for the given entries.

        When no lock is given None is returned as its value.
        """
        redis_keys = [
            self._lock_key(model, identifiers) for identifiers in identifiers_list
        ]
        # When there is no lock return error: ok could abort the processing.
        return self._get_redis_data(redis_keys)

    def wait(
        self,
        model: models.Model,
        identifiers_list: Sequence[Sequence],
        timeout: int = 60,
        refresh_interval: int = 1,
    ) -> set[Optional[RedisLockStatus]]:
        """Wait for the locks for the given entries to be released.

        If the locks are not released within the given timeout proceed anyway.
        """
        redis_lock_keys = [
            self._lock_key(model, identifiers) for identifiers in identifiers_list
        ]
        start_time = time.time()
        while time.time() - start_time < timeout:
            statuses = set(self._get_redis_data(redis_lock_keys))
            if RedisLockStatus.PROCESSING not in statuses:
                break
            time.sleep(refresh_interval)
        return statuses

    def get(
        self,
        model: models.Model,
        identifiers: Identifier,
        field_names: Iterable[str],
    ) -> FieldValues:
        """Obtain the set of fields from the redis cache.

        If field_name is requested that is not cached the rusult for that key is None.

        The query is run in a transaction.
        """

        def get_redis_data(
            pipeline: redis.client.Pipeline,
        ) -> dict[str, Optional[Union[str, datetime]]]:
            """Get the redis data from the specific key."""

            cached_data = dict()
            try:
                unpickled_data = pipeline.get(redis_key)
            except redis.exceptions.RedisError:
                logger.exception(
                    __(
                        "Could not retrieve data from Redis for key: '{}'.",
                        ", ".join(redis_key),
                    )
                )
                raise

            try:
                cached_data = pickle.loads(unpickled_data)  # type: ignore
            except pickle.PickleError:
                logger.exception(
                    __(
                        "Could not deserialize data from Redis for key: '{}'.",
                        redis_key,
                    )
                )
                raise

            return {
                field_name: cached_data.get(field_name) for field_name in field_names
            }

        redis_key = self.get_redis_key(model, identifiers)
        # If there is another command processing the same cache entry wait for it to
        # finish for up to timeout seconds.
        self.wait(model, (identifiers,), timeout=60)
        return self._redis.transaction(
            get_redis_data, redis_key, value_from_callable=True
        )

    def clear(
        self,
        model: Optional[models.Model] = None,
        identifiers: Optional[Identifier] = None,
    ):
        """Clear the entire Redis cache for the given data object.

        When data_id is not given the entire cache is cleared.

        :raises RuntimeError: if identifiers are given without the content type.
        """
        key_prefix = f"{self._get_redis_key_prefix(model, identifiers)}*"
        redis_keys = list(self._redis.scan_iter(key_prefix))
        if redis_keys:
            self._redis.unlink(*redis_keys)

    def mset(
        self,
        model: models.Model,
        to_cache: Cache,
        expiration_time: Optional[int] = None,
    ):
        """Set multiple keys in the redis cache.

        The method does not run in a transaction.

        To speed up the method it is not checked if the field is cached.

        When error occurs it is logged but not raised. Only the part of the given keys
        is cached in such case.

        Note: expiration time is set key by key so it can be slow for large number of
        keys.
        """
        redis_data = {
            self.get_redis_key(model, identifiers): pickle.dumps(item)
            for identifiers, item in to_cache.items()
        }
        # Write data in chunks. Do not abort if single chunk fails.
        for chunk in chunked(redis_data.items(), self.KEY_BATCH_SIZE):
            try:
                self._redis.mset(dict(chunk))  # type: ignore
                if expiration_time:
                    for key, _ in chunk:
                        self._redis.expire(key, expiration_time)
            except redis.exceptions.RedisError:
                logger.exception(
                    __(
                        "Could not set data in Redis for keys: {}",
                        ", ".join(redis_data.keys()),
                    )
                )

    def set(
        self,
        model: models.Model,
        identifiers: Identifier,
        field_values: FieldValues,
        expiration_time: Optional[int] = None,
    ):
        """Set the redis cache in the transaction.

        When field is not in the set of cached fields it is silentry ignored.
        When the given content type/identifiers pair is already in the cache it is updated.
        """

        def update_cache(pipeline: redis.client.Pipeline):
            """Update the cache using pipeline."""
            cached_fields = {key: value for key, value in field_values.items()}
            redis_key = self.get_redis_key(model, identifiers)
            existing_cache = {}
            with suppress(Exception):
                existing_cache = pickle.loads(pipeline.get(redis_key))  # type: ignore
            cache_data = pickle.dumps({**existing_cache, **cached_fields})
            pipeline.set(redis_key, cache_data)
            if expiration_time:
                pipeline.expire(redis_key, expiration_time)

        self._redis.transaction(update_cache)


redis_cache = RedisCache()


class CachedObjectManager(PluginManager["CachedObjectPlugin"]):
    """Redis cache plugin manager."""

    def _get_plugin_identifier(self, model: models.Model) -> str:
        """Get the plugin identifier."""
        return model._meta.label_lower

    def get_plugin(self, model: models.Model) -> "CachedObjectPlugin":
        """Get the plugin based on the content type.

        :raises KeyError: when no plugin is registered for the given content type.
        """
        return self._plugins[self._get_plugin_identifier(model)]

    def mcache(self, instances: models.QuerySet) -> None:
        """Cache the given queryset.

        Items with the same identifiers are overwritten.
        """
        plugin = self.get_plugin(instances.model)
        to_cache = plugin.serialize(instances)
        redis_cache.mset(plugin.model, to_cache, plugin.expiration_time)

    def cache(self, instance: models.Model) -> None:
        """Cache the given instance."""
        plugin = self.get_plugin(type(instance))
        to_cache = plugin.serialize(instance)
        redis_cache.mset(plugin.model, to_cache, plugin.expiration_time)

    def mget(
        self, model: models.Model, identifiers_list: Sequence[Identifier]
    ) -> list[Optional[FieldValues]]:
        """Get the cache values for the given identifiers."""
        plugin = self.get_plugin(model)
        return redis_cache.mget(plugin.model, identifiers_list)

    def get(
        self, model: models.Model, identifiers: Identifier
    ) -> Optional[FieldValues]:
        """Get the cache values for the given instance."""
        return self.mget(model, [identifiers])[0]

    def update_cache(
        self, model: models.Model, identifiers: Identifier, values: FieldValues
    ):
        """Update the given cache with values."""
        plugin = self.get_plugin(model)
        current_cache = self.get(model, identifiers) or {}
        for field in plugin.cached_fields:
            if cached_field_value := values.get(field):
                current_cache[field] = cached_field_value
        to_cache = {identifiers: current_cache}
        redis_cache.mset(plugin.model, to_cache, plugin.expiration_time)

    def is_cached(self, model: models.Model, field_name: str) -> bool:
        """Check if the given field is cached."""
        return field_name in self.get_plugin(model).cached_fields

    def lock(self, model: models.Model, identifiers_list: Sequence[Identifier]):
        """Create locks for the given entries."""
        return redis_cache.lock(model, identifiers_list)

    def unlock(
        self,
        model: models.Model,
        identifiers_list: Sequence[Identifier],
        status: RedisLockStatus = RedisLockStatus.OK,
    ):
        """Unlock locks for the given entries."""
        return redis_cache.unlock(model, identifiers_list, status)

    def wait(
        self,
        model: models.Model,
        identifiers_list: Sequence[Sequence],
        timeout: int = 60,
    ) -> set[Optional[RedisLockStatus]]:
        """Wait for locks to be released for up to 60 seconds."""
        return redis_cache.wait(model, identifiers_list, timeout)


cache_manager = CachedObjectManager()


class CachedObjectPlugin(Plugin):
    """Cache a single ORM object in Redis.

    The object content type and identifiers are used to create the key under which the
    object is stored. The object is stored as a pickle dump of the dictionary
    mapping cached keys to their values.

    Assumptions:
    - the data is safe, pickle is used to create its binary representation.
    """

    abstract = True
    plugin_manager = cache_manager
    model: models.Model
    # The list of cached fields.
    cached_fields: FieldNames
    # The list of identifier fields: must be a subset of cached_fields.
    identifier_fields: FieldNames
    # Default key expiration time is 1 day.
    expiration_time: int = 24 * 3600

    def _get_identifiers(self, values: FieldValues) -> Identifier:
        """Get the identifiers from values."""
        return tuple(values[field] for field in self.identifier_fields)

    def _serialize_instance(self, instance: models.Model) -> Cache:
        """Return the dictionary to store in redis."""
        values = {
            field_name: getattr(instance, field_name, None)
            for field_name in self.cached_fields
        }
        return {self._get_identifiers(values): values}

    def _serialize_queryset(self, instances: models.QuerySet) -> Cache:
        """Return a dictionaryto store in redis.

        The keys are identifiers and values dictionaries.
        """
        return {
            self._get_identifiers(values): values
            for values in instances.values(*self.cached_fields)
        }

    def serialize(self, instances: Union[models.Model, models.QuerySet]) -> Cache:
        """Serialize the given model instance or queryset."""
        serializer = (
            self._serialize_instance
            if isinstance(instances, models.Model)
            else self._serialize_queryset
        )
        return serializer(instances)

    @classmethod
    def get_identifier(cls) -> str:
        """Return the plugin identifier."""
        return cls.model._meta.label_lower


class DataCache(CachedObjectPlugin):
    """Cache the data model in Redis."""

    model = Data
    cached_fields = ("id", "status", "started", "worker__status")
    identifier_fields = ("id",)



"""Overrides for Django settings in tests."""

import copy

from django.conf import settings

from resolwe.storage import settings as storage_settings

# override all FLOW_EXECUTOR settings that are specified in FLOW_EXECUTOR['TEST']
FLOW_EXECUTOR_SETTINGS = copy.deepcopy(getattr(settings, "FLOW_EXECUTOR", {}))
TEST_SETTINGS_OVERRIDES = getattr(settings, "FLOW_EXECUTOR", {}).get("TEST", {})
FLOW_EXECUTOR_SETTINGS.update(TEST_SETTINGS_OVERRIDES)

# override all FLOW_MANAGER settings that are specified in FLOW_MANAGER['TEST']
FLOW_MANAGER_SETTINGS = copy.deepcopy(getattr(settings, "FLOW_MANAGER", {}))
TEST_SETTINGS_OVERRIDES = getattr(settings, "FLOW_MANAGER", {}).get("TEST", {})
FLOW_MANAGER_SETTINGS.update(TEST_SETTINGS_OVERRIDES)

STORAGE_CONNECTORS = copy.deepcopy(storage_settings.STORAGE_CONNECTORS)
for connector_name, connector_settings in STORAGE_CONNECTORS.items():
    connector_settings["config"].update(connector_settings.get("test", {}))

FLOW_VOLUMES = copy.deepcopy(storage_settings.FLOW_VOLUMES)
for volume_name, volume in FLOW_VOLUMES.items():
    volume["config"].update(volume.get("test", {}))



import os.path

import setuptools

# Get the long description from README.
with open("README.rst", "r") as fh:
    long_description = fh.read()

# Get package metadata from '__about__.py' file.
about = {}
base_dir = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(base_dir, "resolwe", "__about__.py"), "r") as fh:
    exec(fh.read(), about)

setuptools.setup(
    name=about["__title__"],
    use_scm_version=True,
    description=about["__summary__"],
    long_description=long_description,
    long_description_content_type="text/x-rst",
    author=about["__author__"],
    author_email=about["__email__"],
    url=about["__url__"],
    license=about["__license__"],
    # Exclude tests from built/installed package.
    packages=setuptools.find_packages(
        exclude=["tests", "tests.*", "*.tests", "*.tests.*"]
    ),
    package_data={
        "resolwe": [
            "flow/executors/requirements.txt",
            "flow/migrations/*.sql",
            "flow/static/flow/*.json",
            "toolkit/processes/**.yml",
            "toolkit/processes/**.py",
            "toolkit/tools/**.py",
        ]
    },
    python_requires=">=3.10, <3.11",
    install_requires=[
        "asgiref~=3.5.2",
        "asteval==0.9.26",
        "async-timeout~=4.0.2",
        "channels~=3.0.4",
        "channels_redis~=3.4.1",
        # Storage requirement for computing hashes.
        "crcmod",
        "kubernetes~=21.7.0",
        "docker~=6.1.1",
        "Django~=3.2.12",
        "djangorestframework~=3.13.1",
        "django-filter~=21.1",
        "django-versionfield~=1.0.2",
        "django-fernet-fields~=0.6",
        "django-priority-batch >=4.0a1, ==4.*",
        "Jinja2~=3.0.3",
        "jsonschema~=4.0.0",
        "plumbum~=1.7.2",
        "psycopg2-binary~=2.9.3",
        "PyYAML~=6.0",
        "redis~=4.1.3",
        "shellescape~=3.8.1",
        "beautifulsoup4~=4.10.0",
        "Sphinx~=4.3.2",
        "wrapt~=1.13.3",
        "pyzmq~=22.3.0",
        "uvloop~=0.17.0",
    ],
    extras_require={
        "storage_s3": [
            "boto3~=1.21.0",
            "crcmod",
        ],
        "storage_gcs": [
            "crcmod",
            "google-cloud-storage~=2.0.0",
        ],
        "docs": [
            "sphinx_rtd_theme",
            "pyasn1>=0.4.8",
        ],
        "package": [
            "twine",
            "wheel",
        ],
        "test": [
            "black>=22.1.0",
            "check-manifest>=0.47",
            "coverage>=6.2.0",
            "flake8>=4.0.1",
            "pydocstyle>=6.1.1",
            "readme_renderer",
            "setuptools_scm",
            "testfixtures>=6.18.3",
            "tblib>=1.7.0",
            "isort>=5.10.1",
        ],
    },
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Web Environment",
        "Framework :: Django",
        "Intended Audience :: Developers",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
        "Topic :: Internet :: WWW/HTTP :: HTTP Servers",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.10",
    ],
    keywords="resolwe dataflow django",
)
